import re


FLOAT_REGEX = re.compile(r"""
(
    [-+]?       # optional plus or minus sign
    [0-9]*
    \.?         # floating point separator
    [0-9]+
)""", re.VERBOSE)


POSITIVE_FLOAT_REGEX = re.compile(r"([0-9]*\.?[0-9]+)", re.VERBOSE)


POSITIVE_FLOAT_LIST_REGEX = re.compile(r"""
(-0\.0)?                        # First float can be -0.0
(
    ({POSITIVE_FLOAT}+)         # positive float
    \s                          # separated by a space
)*
({POSITIVE_FLOAT}+)             # final positive float with no trailing space.
""".format(POSITIVE_FLOAT=POSITIVE_FLOAT_REGEX.pattern), re.VERBOSE)


TIMESERIE_ENTRY_REGEX = re.compile(r"""
(\d+)
(,)
([-+]?[0-9]*\.?[0-9]+)      # digit,float; for example: 5,-3.09
""", re.VERBOSE)


TIMESERIES_REGEX = re.compile(r"""
(
    {timeserie_entry}       # digit,float; for example: 60,-0.5
    \n                      # separated by newlines
)*
(
    {timeserie_entry}       # last entry does not have a newline.
){{1}}
""".format(timeserie_entry=TIMESERIE_ENTRY_REGEX.pattern), re.VERBOSE)



from abc import ABC
from abc import abstractmethod

from geoalchemy2 import functions as geo_func
from geoalchemy2.types import Geometry
from sqlalchemy import func
from sqlalchemy import not_
from sqlalchemy import types
from sqlalchemy.orm.session import Session

from typing import List, NamedTuple


class BaseCheck(ABC):
    """Base class for all checks.

    A Check defines a constraint on a specific column and its table.
    One can validate if the constrain holds using the method `get_invalid()`.
    This method will return a list of rows (as named_tuples) which are invalid.
    """

    def __init__(self, column):
        self.column = column
        self.table = column.table

    @abstractmethod
    def get_invalid(self, session: Session) -> List[NamedTuple]:
        """Return a list of rows (named_tuples) which are invalid.

        What is invalid is defined in the check. Returns an empty list if no
        invalid rows are present.

        :param session: sqlalchemy.orm.session.Session
        :return: list of named_tuples or empty list if there are no invalid
            rows
        """
        pass

    def get_valid(self, session: Session) -> List[NamedTuple]:
        """Return a list of rows (named_tuples) which are valid.

        :param session: sqlalchemy.orm.session.Session
        :return: list of named_tuples or empty list if there are no valid rows
        """
        all_rows = self.to_check(session)
        invalid_row_ids = set([row.id for row in self.get_invalid(session)])
        valid = []
        for row in all_rows:
            if row.id not in invalid_row_ids:
                valid.append(row)
        return valid

    def to_check(self, session):
        """Return a Query object filtering on the rows this check is applied.

        :param session: sqlalchemy.orm.session.Session
        :return: sqlalchemy.Query
        """
        return session.query(self.table)

    def description(self) -> str:
        """Return a string explaining why rows are invalid according to this
        check.

        :return: string
        """
        return "Invalid value in column '%s'" % self.column

    def __repr__(self) -> str:
        return "<%s: %s.%s>" % (
            type(self).__name__,
            self.table.name,
            self.column.name
        )


class GeneralCheck(BaseCheck):
    """Check to specify with an SQL expression what's valid/invalid.

    Either specify what is valid with `criterion_valid` or what is invalid
    with `criterion_invalid`.
    The criterion should be a sqlalchemy.sql.expression.BinaryExpression (https://docs.sqlalchemy.org/en/13/core/sqlelement.html#sqlalchemy.sql.expression.BinaryExpression)  # noqa
    with operators being within `self.table.columns`
    """

    def __init__(self, criterion_invalid=None, criterion_valid=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.criterion_invalid = criterion_invalid
        self.criterion_valid = criterion_valid

    def get_invalid(self, session):
        if self.criterion_invalid is not None:
            q_invalid = self.to_check(session).filter(self.criterion_invalid)
            return q_invalid.all()
        elif self.criterion_valid is not None:
            q_invalid = self.to_check(session).filter(~self.criterion_valid)
            return q_invalid.all()
        else:
            raise AttributeError("No valid/invalid criterion has been specified")

    def description(self):
        if self.criterion_valid is not None:
            condition = self.criterion_valid.compile(
                compile_kwargs={"literal_binds": True}
            )
        elif self.criterion_invalid is not None:
            condition = not_(self.criterion_invalid).compile(
                compile_kwargs={"literal_binds": True}
            )
        else:
            condition = "no condition specified"
        return "'%s'" % condition


class QueryCheck(BaseCheck):
    """Specify a sqlalchemy.orm.Query object to return invalid instances

    Provides more freedom than the GeneralCheck where you need to specify a
    sqlalchemy.sql.expression.BinaryExpression. For example, QueryCheck allows joins
    on multiple tables"""

    def __init__(self, column, invalid, message):
        super().__init__(column)
        self.invalid = invalid
        self.message = message

    def get_invalid(self, session):
        return list(self.invalid.with_session(session))

    def description(self):
        return self.message


class ForeignKeyCheck(BaseCheck):
    """Check all values in `column` are in `reference_column`.

    Null values are ignored."""

    def __init__(self, reference_column, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reference_column = reference_column

    def get_invalid(self, session):
        q_invalid = self.to_check(session)
        invalid_foreign_keys_query = q_invalid.filter(
            self.column.notin_(session.query(self.reference_column)),
            self.column != None,
        )
        return invalid_foreign_keys_query.all()

    def description(self):
        return "Missing foreign key in column %s, expected reference to %s." % (
            self.column,
            self.reference_column,
        )


class UniqueCheck(BaseCheck):
    """Check all values in `column` are unique

    Null values are ignored."""

    def get_invalid(self, session):
        duplicate_values = (
            session.query(self.column)
            .group_by(self.column)
            .having(func.count(self.column) > 1)
        )
        q_invalid = self.to_check(session)
        invalid_uniques_query = q_invalid.filter(
            self.column.in_(duplicate_values)
        )
        return invalid_uniques_query.all()

    def description(self):
        return "Value in %s should to be unique" % self.column


class NotNullCheck(BaseCheck):
    """"Check all values in `column` that are not null"""

    def get_invalid(self, session):
        q_invalid = self.to_check(session)
        not_null_query = q_invalid.filter(self.column == None)
        return not_null_query.all()

    def description(self):
        return "Value in %s cannot be null" % self.column


class TypeCheck(BaseCheck):
    """Check all values in `column` that are of the column defined type.

    Null values are ignored."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.expected_type = _sqlalchemy_to_sqlite_type(self.column.type)

    def get_invalid(self, session):
        if "sqlite" not in session.bind.dialect.dialect_description:
            return []
        q_invalid = self.to_check(session)
        invalid_type_query = q_invalid.filter(
            func.typeof(self.column) != self.expected_type,
            func.typeof(self.column) != "null",
        )
        return invalid_type_query.all()

    def description(self):
        return "Value in %s should to be of type %s" % (
            self.column,
            self.expected_type
        )


def _sqlalchemy_to_sqlite_type(column_type):
    """Convert the sqlalchemy column type to sqlite data type

    Returns the value similar as the sqlite 'typeof' function.
    Raises TypeError if the column type is unknown.
    See https://www.sqlite.org/datatype3.html

    :param column_type: sqlalchemy.column
    :return: (str)
    """
    if isinstance(column_type, types.TypeDecorator):
        column_type = column_type.impl

    if isinstance(column_type, types.String):
        return "text"
    elif isinstance(column_type, types.Float):
        return "real"
    elif isinstance(column_type, types.Integer):
        return "integer"
    elif isinstance(column_type, types.Boolean):
        return "integer"
    elif isinstance(column_type, types.Numeric):
        return "numeric"
    elif isinstance(column_type, types.Date):
        return "text"
    elif isinstance(column_type, Geometry):
        return "blob"
    elif isinstance(column_type, types.TIMESTAMP):
        return "text"
    else:
        raise TypeError("Unknown column type: %s" % column_type)


class GeometryCheck(BaseCheck):
    """Check all values in `column` are a valid geometry.

    Null values are ignored."""

    def get_invalid(self, session):
        q_invalid = self.to_check(session)
        invalid_geometries = q_invalid.filter(
            geo_func.ST_IsValid(self.column) != True, self.column != None
        )
        return invalid_geometries.all()

    def description(self):
        return "Value in %s is invalid geometry" % self.column


class GeometryTypeCheck(BaseCheck):
    """Check all values in `column` are of geometry type in defined in
    `column`.

    Null values are ignored"""

    def get_invalid(self, session):
        expected_geometry_type = _get_geometry_type(
            self.column, dialect=session.bind.dialect.name
        )
        q_invalid = self.to_check(session)
        invalid_geometry_types_q = q_invalid.filter(
            geo_func.ST_GeometryType(self.column) != expected_geometry_type,
            self.column != None,
        )
        return invalid_geometry_types_q.all()

    def description(self):
        return "Value in %s has invalid geometry type, expected geometry " \
               "type %s" % (self.column, self.column.type.geometry_type)


def _get_geometry_type(column, dialect):
    if dialect == "sqlite":
        return column.type.geometry_type
    elif dialect == "postgresql":
        geom_type = column.type.geometry_type.capitalize()
        return "ST_%s" % geom_type
    else:
        raise TypeError("Unexpected dialect %s" % dialect)


class EnumCheck(BaseCheck):
    """Check all values in `column` are within the defined Enum values of
    `column`.

    Unexpected values are values not defined by its enum_class.

    Null values are ignored"""

    def get_invalid(self, session):
        q_invalid = self.to_check(session)
        invalid_values_q = q_invalid.filter(
            self.column.notin_(list(self.column.type.enum_class))
        )
        return invalid_values_q.all()

    def description(self):
        return (
            "Value in %s has invalid value, expected one of the "
            "following values %s" % (
                self.column,
                list(self.column.type.enum_class)
            )
        )



from .checks.base import BaseCheck
from .threedi_database import ThreediDatabase
from .schema import ModelSchema
from .config import Config

from typing import Iterator, Tuple, NamedTuple


class ThreediModelChecker:
    def __init__(self, threedi_db: ThreediDatabase):
        self.db = threedi_db
        self.schema = ModelSchema(self.db)
        self.schema.validate_schema()
        self.config = Config(self.models)

    @property
    def models(self):
        """Returns a list of declared models"""
        return self.schema.declared_models

    def errors(self) -> Iterator[Tuple[BaseCheck, NamedTuple]]:
        """Iterates and applies checks, returning any failing rows.

        :return: Tuple of the applied check and the failing row.
        """
        session = self.db.get_session()
        for check in self.checks():
            model_errors = check.get_invalid(session)
            for error_row in model_errors:
                yield check, error_row

    def checks(self) -> Iterator[BaseCheck]:
        """Iterates over all configured checks

        :return: implementations of BaseChecks
        """
        for check in self.config.checks:
            yield check

    def check_table(self, table):
        pass

    def check_column(self, column):
        pass



# Copied from ThreeDiToolbox
from builtins import object
import copy

from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine
from sqlalchemy.event import listen
from sqlalchemy.ext.declarative import declarative_base


Base = declarative_base()


def load_spatialite(con, connection_record):
    """Load spatialite extension as described in
    https://geoalchemy-2.readthedocs.io/en/latest/spatialite_tutorial.html"""
    import sqlite3

    con.enable_load_extension(True)
    cur = con.cursor()
    libs = [
        # SpatiaLite >= 4.2 and Sqlite >= 3.7.17, should work on all platforms
        ("mod_spatialite", "sqlite3_modspatialite_init"),
        # SpatiaLite >= 4.2 and Sqlite < 3.7.17 (Travis)
        ("mod_spatialite.so", "sqlite3_modspatialite_init"),
        # SpatiaLite < 4.2 (linux)
        ("libspatialite.so", "sqlite3_extension_init"),
    ]
    found = False
    for lib, entry_point in libs:
        try:
            cur.execute("select load_extension('{}', '{}')".format(lib, entry_point))
        except sqlite3.OperationalError:
            continue
        else:
            found = True
            break
    if not found:
        raise RuntimeError("Cannot find any suitable spatialite module")
    cur.close()
    con.enable_load_extension(False)


class ThreediDatabase(object):
    def __init__(self, connection_settings, db_type="spatialite", echo=False):
        """

        :param connection_settings:
        db_type (str choice): database type. 'sqlite' and 'postgresql' are
                              supported
        """
        self.settings = connection_settings
        self.db_type = db_type
        self.echo = echo

        self._engine = None
        self._base_metadata = None

    @property
    def engine(self):
        return self.get_engine()

    def get_engine(self, get_seperate_engine=False):

        if self._engine is None or get_seperate_engine:
            if self.db_type == "spatialite":
                engine = create_engine(
                    "sqlite:///{0}".format(self.settings["db_path"]), echo=self.echo
                )
                listen(engine, "connect", load_spatialite)
                if get_seperate_engine:
                    return engine
                else:
                    self._engine = engine

            elif self.db_type == "postgres":
                con = (
                    "postgresql://{username}:{password}@{host}:"
                    "{port}/{database}".format(**self.settings)
                )

                engine = create_engine(con, echo=self.echo)
                if get_seperate_engine:
                    return engine
                else:
                    self._engine = engine

        return self._engine

    def get_metadata(self, including_existing_tables=True, engine=None):

        if including_existing_tables:
            metadata = copy.deepcopy(Base.metadata)
            if engine is None:
                engine = self.engine

            metadata.bind = engine
            metadata.reflect(extend_existing=True)
            return metadata
        else:
            if self._base_metadata is None:
                self._base_metadata = copy.deepcopy(Base.metadata)
            return self._base_metadata

    def get_session(self):
        return sessionmaker(bind=self.engine)()

    def check_connection(self):
        """Check if there a connection can be started with the database

        :return: True if a connection can be established, otherwise raises an
            appropriate error.
        """
        session = self.get_session()
        r = session.execute("select 1")
        return r.fetchone()



from collections import Counter


def print_errors(errors):
    """Simply prints all errors to stdout

    :param errors: iterator of BaseModelError
    """
    for error in errors:
        print(error)


def export_to_file(errors, file):
    """Write errors to a new file, separated with newlines.

    File cannot be an already existing file.

    :param errors: iterator of BaseModelError
    :param file:
    :return: None
    :raise FileExistsError: if the file already exists
    """
    with open(file, "w") as f:
        for error in errors:
            f.write(str(error) + "\n")


def summarize_type_errors(errors):
    """Return an summary of the errors, aggregated on the type of errors

    Count for each type of error is returned.

    :param errors: iterator of BaseModelError
    :return: dict
    """

    def _get_error_type(errors):
        for error in errors:
            yield type(error).__name__

    summary = Counter(_get_error_type(errors))
    return summary, sum(summary.values())


def summarize_column_errors(errors):
    """Return a summary of the errors, aggregated on columns

    For each column the number of errors are returned. Columns with no errors
    are not returned.
    """

    def _get_error_table_column(errors):
        for error in errors:
            yield "%s.%s" % (error.column.table.name, error.column.name)

    summary = Counter(_get_error_table_column(errors))
    return summary, sum(summary.values())


def format_check_results(check, invalid_row):
    OUTPUT_FORMAT = '{check!r} row: {row_id:d} value: "{row_value!s}" {description!s}'
    return OUTPUT_FORMAT.format(
        check=check,
        row_id=invalid_row.id,
        row_value=getattr(invalid_row, check.column.name),
        description=check.description(),
    )



from geoalchemy2 import functions as geo_func
from sqlalchemy import func
from sqlalchemy.orm import aliased, Query

from threedi_modelchecker.checks import patterns
from .base import BaseCheck
from ..threedi_model import constants
from ..threedi_model import models


class BankLevelCheck(BaseCheck):
    """Check 'CrossSectionLocation.bank_level' is not null if
    calculation_type is CONNECTED or DOUBLE_CONNECTED.
    """

    def __init__(self):
        super().__init__(column=models.CrossSectionLocation.bank_level)

    def get_invalid(self, session):
        q = session.query(self.table).filter(
            models.CrossSectionLocation.bank_level == None,
            models.CrossSectionLocation.channel.has(
                models.Channel.calculation_type.in_(
                    [
                        constants.CalculationType.CONNECTED,
                        constants.CalculationType.DOUBLE_CONNECTED,
                    ]
                )
            ),
        )
        return q.all()

    def description(self):
        return "CrossSectionLoaction.Banklevel cannot be null when calculation_type " \
               "is CONNECTED or DOUBLE_CONNECTED"


class CrossSectionShapeCheck(BaseCheck):
    """Check if all CrossSectionDefinition.shape are valid"""

    def __init__(self):
        super().__init__(column=models.CrossSectionDefinition.shape)

    def get_invalid(self, session):
        cross_section_definitions = session.query(self.table)
        invalid_cross_section_shapes = []

        for cross_section_definition in cross_section_definitions.all():
            shape = cross_section_definition.shape
            width = cross_section_definition.width
            height = cross_section_definition.height
            if shape == constants.CrossSectionShape.RECTANGLE:
                if not valid_rectangle(width, height):
                    invalid_cross_section_shapes.append(cross_section_definition)
            elif shape == constants.CrossSectionShape.CIRCLE:
                if not valid_circle(width):
                    invalid_cross_section_shapes.append(cross_section_definition)
            elif shape == constants.CrossSectionShape.EGG:
                if not valid_egg(width):
                    invalid_cross_section_shapes.append(cross_section_definition)
            if shape == constants.CrossSectionShape.TABULATED_RECTANGLE:
                if not valid_tabulated_shape(width, height):
                    invalid_cross_section_shapes.append(cross_section_definition)
            elif shape == constants.CrossSectionShape.TABULATED_TRAPEZIUM:
                if not valid_tabulated_shape(width, height):
                    invalid_cross_section_shapes.append(cross_section_definition)
        return invalid_cross_section_shapes

    def description(self):
        return "Invalid CrossSectionShape"


def valid_rectangle(width, height):
    if width is None:  # width is required
        return False
    width_match = patterns.POSITIVE_FLOAT_REGEX.fullmatch(width)
    if height is not None:  # height is not required
        height_match = patterns.POSITIVE_FLOAT_REGEX.fullmatch(height)
    else:
        height_match = True
    return width_match and height_match


def valid_circle(width):
    if width is None:
        return False
    return patterns.POSITIVE_FLOAT_REGEX.fullmatch(width)


def valid_egg(width):
    if width is None:
        return False
    try:
        w = float(width)
    except ValueError:
        return False
    return w > 0


def valid_tabulated_shape(width, height):
    """Return if the tabulated shape is valid.

    Validating that the strings `width` and `height` contain positive floats
    was previously done using a regex. However, experiments showed that
    trying to split the string and reading in the floats is much faster.

    :param width: string of widths
    :param height: string of heights
    :return: True if the shape if valid
    """
    if width is None or height is None:
        return False
    height_string_list = height.split(" ")
    width_string_list = width.split(" ")
    if len(height_string_list) != len(width_string_list):
        return False
    try:
        # first height must be 0
        first_height = float(height_string_list[0])
        if first_height != 0:
            return False
    except ValueError:
        return False
    previous_height = -1
    for h_string, w_string in zip(height_string_list, width_string_list):
        try:
            h = float(h_string)
            w = float(w_string)
        except ValueError:
            return False
        if h < 0:
            return False
        if w < 0:
            return False
        if h < previous_height:
            # height must be increasing
            return False
        previous_height = h
    return True


class TimeseriesCheck(BaseCheck):
    """Check that `column` has the time series pattern: digit,float\n

    The first digit is the timestep in minutes, the float is a value depending
    on the type of timeseries.

    Example of a timeserie: 0,-0.5\n59,-0.5\n60,-0.5\n61,-0.5\n9999,-0.5

    All timeseries in the table should contain the same timesteps.
    """

    def get_invalid(self, session):
        invalid_timeseries = []
        required_timesteps = {}
        rows = session.query(self.table).all()

        for row in rows:
            timeserie = row.timeseries
            if not patterns.TIMESERIES_REGEX.fullmatch(timeserie):
                invalid_timeseries.append(row)
                continue

            timesteps = {
                time for time, *_ in patterns.TIMESERIE_ENTRY_REGEX.findall(timeserie)
            }
            if not required_timesteps:
                # Assume the first timeserie defines the required timesteps.
                # All others should have the same timesteps.
                required_timesteps = timesteps
                continue
            if timesteps != required_timesteps:
                invalid_timeseries.append(row)

        return invalid_timeseries

    def description(self):
        return "Invalid timeseries"


class Use0DFlowCheck(BaseCheck):
    """Check that when use_0d_flow in global settings is configured to 1 or to
    2, there is at least one impervious surface or surfaces respectively.
    """

    def __init__(self):
        super().__init__(column=models.GlobalSetting.use_0d_inflow)

    def to_check(self, session):
        """Return a Query object on which this check is applied"""
        return session.query(models.GlobalSetting).filter(
            models.GlobalSetting.use_0d_inflow != 0
        )

    def get_invalid(self, session):
        surface_count = session.query(func.count(models.Surface.id)).scalar()
        impervious_surface_count = session.query(
            func.count(models.ImperviousSurface.id)
        ).scalar()

        invalid_rows = []
        for row in self.to_check(session):
            if row.use_0d_inflow == 1 and impervious_surface_count == 0:
                invalid_rows.append(row)
            elif row.use_0d_inflow == 2 and surface_count == 0:
                invalid_rows.append(row)
            else:
                continue
        return invalid_rows

    def description(self):
        return (
            "When %s is used, there should exists at least one "
            "(impervious) surface." % self.column
        )


class ConnectionNodes(BaseCheck):
    """Check that all connection nodes are connected to at least one of the
    following objects:
    - Culvert
    - Channel
    - Pipe
    - Orifice
    - Pumpstation
    - Weir
    """

    def __init__(self):
        super().__init__(column=models.ConnectionNode.id)

    def get_invalid(self, session):
        raise NotImplementedError


class ConnectionNodesLength(BaseCheck):
    """Check that the distance between `start_node` and `end_node` is at least
    `min_distance`. The coords will be transformed into (the first entry) of
    GlobalSettings.epsg_code. The `min_distance` will be interpreted as these units.
    For example epsg:28992 will be in meters while epsg:4326 is in degrees."""

    def __init__(self, start_node, end_node, min_distance: float, *args, **kwargs):
        """

        :param start_node: column name of the start node
        :param end_node: column name of the end node
        :param min_distance: minimum required distance between start and end node
        """
        super().__init__(*args, **kwargs)
        self.start_node = start_node
        self.end_node = end_node
        self.min_distance = min_distance

    def get_invalid(self, session):
        start_node = aliased(models.ConnectionNode)
        end_node = aliased(models.ConnectionNode)
        q = Query(
            self.column.class_
        ).join(
            start_node, self.start_node
        ).join(
            end_node, self.end_node
        ).filter(
            geo_func.ST_Distance(
                geo_func.ST_Transform(
                    start_node.the_geom,
                    Query(models.GlobalSetting.epsg_code).limit(1).label('epsg_code')
                ),
                geo_func.ST_Transform(
                    end_node.the_geom,
                    Query(models.GlobalSetting.epsg_code).limit(1).label('epsg_code')
                )
            ) < self.min_distance
        )
        return list(q.with_session(session).all())

    def description(self) -> str:
        return f"The distance of start- and end connectionnode of {self.table} is " \
               f"too short, should be at least {self.min_distance} meters"



import click

from threedi_modelchecker.model_checks import ThreediModelChecker
from threedi_modelchecker.threedi_database import ThreediDatabase
from threedi_modelchecker import exporters


@click.group()
@click.option("-f", "--file", help="Write errors to file, instead of stdout")
@click.option(
    "-s", "--sum/--no-sum", default=False, help="Prints a summary instead of all errors"
)
@click.pass_context
def check_model(ctx, file, sum):
    """Checks the threedi-model for errors"""
    click.echo("Parsing threedi-model for any errors")
    if file:
        click.echo("Model errors will be written to %s" % file)
    if sum:
        click.echo("Printing a summary of the found errors")


@check_model.command()
@click.option("-d", "--database", required=True, help="database name to connect to")
@click.option("-h", "--host", required=True, help="database server host")
@click.option("-p", "--port", required=True, default=5432, help="database server port")
@click.option("-u", "--username", required=True, help="database username")
@click.pass_context
def postgis(context, database, host, port, username, password):
    """Parse a postgis model"""
    postgis_settings = {
        "host": host,
        "port": port,
        "database": database,
        "username": username,
        "password": password,
    }
    db = ThreediDatabase(
        connection_settings=postgis_settings, db_type="postgres", echo=False
    )
    process(db, context.parent)


@check_model.command()
@click.option(
    "-s",
    "--sqlite",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="sqlite file",
)
@click.pass_context
def sqlite(context, sqlite):
    """Parse a sqlite model"""
    sqlite_settings = {"db_path": sqlite, "db_file": sqlite}
    db = ThreediDatabase(
        connection_settings=sqlite_settings, db_type="spatialite", echo=False
    )
    process(db, context.parent)


def process(threedi_db, context):
    mc = ThreediModelChecker(threedi_db)
    model_errors = mc.errors()

    file_output = context.params.get("file")
    summary = context.params.get("sum")
    if context.params.get("file"):
        exporters.export_to_file(model_errors, file_output)
    if summary:
        error_summary, total_errors = exporters.summarize_type_errors(model_errors)
        click.echo("---SUMMARY---")
        click.echo("Total number of errors: %s" % total_errors)
        click.echo(error_summary)
    if not summary and not file_output:
        exporters.print_errors(model_errors)

    click.echo("Finished processing model")


if __name__ == "__main__":
    exit(check_model())



class MigrationMissingError(Exception):
    """Raised when 3Di model is missing migrations."""
    pass


class MigrationTooHighError(Exception):
    """Raised when 3Di model has applied more migrations than expected.

    The expected migration id and name for the 3Di model is configured in
    threedi_modelchecker.threedi_model.constants in LATEST_MIGRATION_ID and
    LATEST_MIGRATION_NAME.
    """
    pass


class MigrationNameError(Exception):
    """Raised when last migration name is unexpected"""
    pass



from setuptools import setup

version = '0.10.1'

long_description = "\n\n".join([open("README.rst").read()])

install_requires = [
    "Click",
    "GeoAlchemy2>=0.6",
    "SQLAlchemy>=0.8",
]

tests_require = [
    "factory_boy",
    "pytest",
    "mock",
    "pytest-cov"
]

setup(
    name="threedi-modelchecker",
    version=version,
    description="Checks validity of a threedi-model",
    long_description=long_description,
    # Get strings from http://www.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Development Status :: 4 - Beta",
    ],
    keywords=[],
    author="Richard Boon",
    author_email="richard.boon@nelen-schuurmans.nl",
    url="https://github.com/nens/threedi-modelchecker",
    license="MIT",
    packages=["threedi_modelchecker"],
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={"test": tests_require},
    entry_points={
        "console_scripts": [
            "threedi-modelchecker = threedi_modelchecker.scripts:check_model"
        ]
    },
)



from sqlalchemy import func

from .threedi_model import constants
from .threedi_model import models
from .errors import MigrationMissingError
from .errors import MigrationTooHighError  # noqa
from .errors import MigrationNameError  # noqa


class ModelSchema:
    def __init__(self, threedi_db, declared_models=models.DECLARED_MODELS):
        self.db = threedi_db
        self.declared_models = declared_models

    def _latest_migration(self):
        """Returns a tuple with latest migration id and name"""
        session = self.db.get_session()
        latest_migration_id = session.query(
            func.max(models.SouthMigrationHistory.id)
        ).scalar()
        latest_migration_name = (
            session.query(models.SouthMigrationHistory.migration)
            .filter(models.SouthMigrationHistory.id == latest_migration_id)
            .scalar()
        )
        return latest_migration_id, latest_migration_name

    def validate_schema(self):
        """Very basic validation of 3Di schema.

        Check that the database has the latest migration applied. If the
        latest migrations is applied, we assume the database also contains all
        tables and columns defined in threedi_model.models.py.

        :return: True if the threedi_db schema is valid, raises an error otherwise.
        :raise MigrationMissingError, MigrationTooHighError, MigrationNameError
        """
        migration_id, migration_name = self._latest_migration()
        if migration_id is None or migration_id < constants.LATEST_MIGRATION_ID:
            raise MigrationMissingError
        # TODO: on 08-07-2019 3Di has been released with the newest migration
        #  174: '0172_auto__del_v2initialwaterlevel__del_field_v2orifice_max_capacity__del_f'.  # noqa
        #  This migrations deletes some fields, which were already not present in the
        #  models of the threedi-modelchecker. Therefore we can both accept migration
        #  173 and 174.
        #  Because inpy has not been run over all the existing models during this new
        #  release, models downloaded via 3id.lizard.net/models still have the
        #  migration 173 (unless a user has specifically re-run inpy for the model).
        elif migration_id > constants.LATEST_MIGRATION_ID:
            # don't raise warning for now, see comments above.
            # raise MigrationTooHighError
            pass
        elif migration_name != constants.LATEST_MIGRATION_NAME:
            # don't fix on a specific migration name, see comments above.
            # raise MigrationNameError
            pass
        return (
            migration_id == constants.LATEST_MIGRATION_ID
            and migration_name == constants.LATEST_MIGRATION_NAME
        )

    def get_missing_tables(self):
        pass

    def get_missing_columns(self):
        pass



from typing import List

from geoalchemy2 import functions as geo_func
from sqlalchemy import Integer, func
from sqlalchemy import and_
from sqlalchemy import cast
from sqlalchemy import or_
from sqlalchemy.orm import Query

from .checks.base import QueryCheck, ForeignKeyCheck, UniqueCheck, \
    TypeCheck, GeometryCheck, GeometryTypeCheck, EnumCheck, BaseCheck
from .checks.base import GeneralCheck
from .checks.factories import generate_enum_checks
from .checks.factories import generate_foreign_key_checks
from .checks.factories import generate_geometry_checks
from .checks.factories import generate_geometry_type_checks
from .checks.factories import generate_not_null_checks
from .checks.factories import generate_type_checks
from .checks.factories import generate_unique_checks
from .checks.other import BankLevelCheck, CrossSectionShapeCheck, ConnectionNodesLength
from .checks.other import TimeseriesCheck
from .checks.other import Use0DFlowCheck
from .threedi_model import models
from .threedi_model.models import constants

FOREIGN_KEY_CHECKS: List[ForeignKeyCheck] = []
UNIQUE_CHECKS: List[UniqueCheck] = []
INVALID_TYPE_CHECKS: List[TypeCheck] = []
INVALID_GEOMETRY_CHECKS: List[GeometryCheck] = []
INVALID_GEOMETRY_TYPE_CHECKS: List[GeometryTypeCheck] = []
INVALID_ENUM_CHECKS: List[EnumCheck] = []

TIMESERIES_CHECKS: List[TimeseriesCheck] = [
    TimeseriesCheck(models.BoundaryCondition1D.timeseries),
    TimeseriesCheck(models.BoundaryConditions2D.timeseries),
    TimeseriesCheck(models.Lateral1d.timeseries),
    TimeseriesCheck(models.Lateral2D.timeseries),
]

RANGE_CHECKS: List[BaseCheck] = [
    GeneralCheck(
        column=models.CrossSectionLocation.friction_value,
        criterion_valid=models.CrossSectionLocation.friction_value > 0,
    ),
    GeneralCheck(
        column=models.Culvert.friction_value,
        criterion_valid=models.Culvert.friction_value >= 0,
    ),
    GeneralCheck(
        column=models.GroundWater.phreatic_storage_capacity,
        criterion_valid=and_(
            models.GroundWater.phreatic_storage_capacity >= 0,
            models.GroundWater.phreatic_storage_capacity <= 1
        ),
    ),
    GeneralCheck(
        column=models.ImperviousSurface.area,
        criterion_valid=models.ImperviousSurface.area >= 0,
    ),
    GeneralCheck(
        column=models.ImperviousSurface.dry_weather_flow,
        criterion_valid=models.ImperviousSurface.dry_weather_flow >= 0,
    ),
    GeneralCheck(
        column=models.ImperviousSurfaceMap.percentage,
        criterion_valid=models.ImperviousSurfaceMap.percentage >= 0,
    ),
    GeneralCheck(
        column=models.Interflow.porosity,
        criterion_valid=and_(
            models.Interflow.porosity >= 0,
            models.Interflow.porosity <= 1,
        ),
    ),
    GeneralCheck(
        column=models.Interflow.impervious_layer_elevation,
        criterion_valid=models.Interflow.impervious_layer_elevation >= 0,
    ),
    GeneralCheck(
        column=models.Orifice.discharge_coefficient_negative,
        criterion_valid=models.Orifice.discharge_coefficient_negative >= 0,
    ),
    GeneralCheck(
        column=models.Orifice.discharge_coefficient_positive,
        criterion_valid=models.Orifice.discharge_coefficient_positive >= 0,
    ),
    GeneralCheck(
        column=models.Orifice.friction_value,
        criterion_valid=models.Orifice.friction_value >= 0,
    ),
    GeneralCheck(
        column=models.Pipe.dist_calc_points,
        criterion_valid=models.Pipe.dist_calc_points > 0,
    ),
    GeneralCheck(
        column=models.Pipe.friction_value,
        criterion_valid=models.Pipe.friction_value >= 0,
    ),
    GeneralCheck(
        column=models.Pumpstation.upper_stop_level,
        criterion_valid=and_(
            models.Pumpstation.upper_stop_level > models.Pumpstation.lower_stop_level,
            models.Pumpstation.upper_stop_level > models.Pumpstation.start_level,
        )
    ),
    GeneralCheck(
        column=models.Pumpstation.lower_stop_level,
        criterion_valid=and_(
            models.Pumpstation.lower_stop_level < models.Pumpstation.start_level,
            models.Pumpstation.lower_stop_level < models.Pumpstation.upper_stop_level,
        )
    ),
    GeneralCheck(
        column=models.Pumpstation.start_level,
        criterion_valid=and_(
            models.Pumpstation.start_level > models.Pumpstation.lower_stop_level,
            models.Pumpstation.start_level < models.Pumpstation.upper_stop_level,
        )
    ),
    GeneralCheck(
        column=models.Pumpstation.capacity,
        criterion_valid=models.Pumpstation.capacity >= 0,
    ),
    GeneralCheck(
        column=models.SimpleInfiltration.infiltration_rate,
        criterion_valid=models.SimpleInfiltration.infiltration_rate >= 0,
    ),
    GeneralCheck(
        column=models.Surface.nr_of_inhabitants,
        criterion_valid=models.Surface.nr_of_inhabitants >= 0,
    ),
    GeneralCheck(
        column=models.Surface.dry_weather_flow,
        criterion_valid=models.Surface.dry_weather_flow >= 0,
    ),
    GeneralCheck(
        column=models.Surface.area,
        criterion_valid=models.Surface.area >= 0,
    ),
    GeneralCheck(
        column=models.SurfaceMap.percentage,
        criterion_valid=and_(
            models.SurfaceMap.percentage >= 0,
            models.SurfaceMap.percentage <= 100,
        ),
    ),
    GeneralCheck(
        column=models.SurfaceParameter.outflow_delay,
        criterion_valid=models.SurfaceParameter.outflow_delay >= 0,
    ),
    GeneralCheck(
        column=models.SurfaceParameter.max_infiltration_capacity,
        criterion_valid=models.SurfaceParameter.max_infiltration_capacity >= 0,
    ),
    GeneralCheck(
        column=models.SurfaceParameter.min_infiltration_capacity,
        criterion_valid=models.SurfaceParameter.min_infiltration_capacity >= 0,
    ),
    GeneralCheck(
        column=models.SurfaceParameter.infiltration_decay_constant,
        criterion_valid=models.SurfaceParameter.infiltration_decay_constant >= 0,
    ),
    GeneralCheck(
        column=models.SurfaceParameter.infiltration_recovery_constant,
        criterion_valid=models.SurfaceParameter.infiltration_recovery_constant >= 0,
    ),
    GeneralCheck(
        column=models.Weir.discharge_coefficient_negative,
        criterion_valid=models.Weir.discharge_coefficient_negative >= 0,
    ),
    GeneralCheck(
        column=models.Weir.discharge_coefficient_positive,
        criterion_valid=models.Weir.discharge_coefficient_positive >= 0,
    ),
    GeneralCheck(
        column=models.Weir.friction_value,
        criterion_valid=models.Weir.friction_value >= 0,
    ),
    GeneralCheck(
        column=models.GlobalSetting.maximum_sim_time_step,
        criterion_valid=models.GlobalSetting.maximum_sim_time_step >= models.GlobalSetting.sim_time_step,  # noqa: E501
    ),
    GeneralCheck(
        column=models.GlobalSetting.sim_time_step,
        criterion_valid=models.GlobalSetting.sim_time_step >= models.GlobalSetting.minimum_sim_time_step,  # noqa: E501
    ),
]

OTHER_CHECKS: List[BaseCheck] = [
    BankLevelCheck(),
    CrossSectionShapeCheck(),
    # 1d boundary conditions cannot be connected to a pumpstation
    GeneralCheck(
        column=models.BoundaryCondition1D.connection_node_id,
        criterion_invalid=or_(
            models.BoundaryCondition1D.connection_node_id == models.Pumpstation.connection_node_start_id,  # noqa: E501
            models.BoundaryCondition1D.connection_node_id == models.Pumpstation.connection_node_end_id,  # noqa: E501
        )
    ),
    GeneralCheck(
        column=models.GlobalSetting.nr_timesteps,
        criterion_valid=cast(models.GlobalSetting.output_time_step, Integer)
        % cast(models.GlobalSetting.sim_time_step, Integer) == 0
    ),
    Use0DFlowCheck()
]


CONDITIONAL_CHECKS = [
    QueryCheck(
        column=models.ConnectionNode.storage_area,
        invalid=Query(models.ConnectionNode).join(
            models.Manhole
        ).filter(
            models.ConnectionNode.storage_area <= 0
        ),
        message="The ConnectionNode.storage_area > 0 "
                "when the ConnectionNode is a Manhole"
    ),
    QueryCheck(
        column=models.CrossSectionLocation.reference_level,
        invalid=Query(models.CrossSectionLocation).filter(
            models.CrossSectionLocation.bank_level != None,
            models.CrossSectionLocation.reference_level >= models.CrossSectionLocation.bank_level  # noqa: E501
        ),
        message="CrossSectionLocation.reference_level < CrossSectionLocation.bank_level"
                "when CrossSectionLocation.bank_level is not null"
    ),
    QueryCheck(
        column=models.GlobalSetting.initial_groundwater_level_type,
        invalid=Query(models.GlobalSetting).filter(
            models.GlobalSetting.initial_groundwater_level_type == None,
            or_(
                models.GlobalSetting.initial_groundwater_level_file != None,
                models.GlobalSetting.initial_groundwater_level != None
            )
        ),
        message="GlobalSetting.initial_groundwater_level_type cannot be null when "
                "GlobalSetting.initial_groundwater_level_file is not null or "
                "GlobalSetting.initial_groundwater_level is not null"
    ),
    QueryCheck(
        column=models.GlobalSetting.water_level_ini_type,
        invalid=Query(models.GlobalSetting).filter(
            models.GlobalSetting.water_level_ini_type == None,
            models.GlobalSetting.initial_waterlevel_file != None
        ),
        message="GlobalSetting.water_level_ini_type cannot be null when "
                "GlobalSetting.initial_waterlevel_file is not null"
    ),
    QueryCheck(
        column=models.GlobalSetting.dem_obstacle_height,
        invalid=Query(models.GlobalSetting).filter(
            models.GlobalSetting.dem_obstacle_height <= 0,
            models.GlobalSetting.dem_obstacle_detection == True
        ),
        message="GlobalSetting.dem_obstacle_height > 0 when "
                "GlobalSetting.dem_obstacle_detection == True"
    ),
    QueryCheck(
        column=models.GroundWater.equilibrium_infiltration_rate_type,
        invalid=Query(models.GroundWater).filter(
            models.GroundWater.equilibrium_infiltration_rate_type == None,
            models.GroundWater.equilibrium_infiltration_rate_file != None
        ),
        message="GroundWater.equilibrium_infiltration_rate_type cannot be null when "
                "GroundWater.equilibrium_infiltration_rate_file is not null"
    ),
    QueryCheck(
        column=models.GroundWater.infiltration_decay_period_type,
        invalid=Query(models.GroundWater).filter(
            models.GroundWater.infiltration_decay_period_type == None,
            models.GroundWater.infiltration_decay_period_type != None
        ),
        message="GroundWater.infiltration_decay_period_type cannot be null when "
                "GroundWater.infiltration_decay_period_type is not null"
    ),
    QueryCheck(
        column=models.GroundWater.groundwater_hydro_connectivity_type,
        invalid=Query(models.GroundWater).filter(
            models.GroundWater.groundwater_hydro_connectivity_type == None,
            models.GroundWater.groundwater_hydro_connectivity_file != None
        ),
        message="GroundWater.groundwater_hydro_connectivity_type cannot be null when "
                "GroundWater.groundwater_hydro_connectivity_file is not null"
    ),
    QueryCheck(
        column=models.GroundWater.groundwater_impervious_layer_level_type,
        invalid=Query(models.GroundWater).filter(
            models.GroundWater.groundwater_impervious_layer_level_type == None,
            models.GroundWater.groundwater_impervious_layer_level_file != None
        ),
        message="GroundWater.groundwater_impervious_layer_level_type cannot be null "
                "when GroundWater.groundwater_impervious_layer_level_file is not null"
    ),
    QueryCheck(
        column=models.GroundWater.initial_infiltration_rate_type,
        invalid=Query(models.GroundWater).filter(
            models.GroundWater.initial_infiltration_rate_type == None,
            models.GroundWater.initial_infiltration_rate_file != None
        ),
        message="GroundWater.initial_infiltration_rate_type cannot be null when "
                "GroundWater.initial_infiltration_rate_file is not null"
    ),
    QueryCheck(
        column=models.GroundWater.phreatic_storage_capacity_type,
        invalid=Query(models.GroundWater).filter(
            models.GroundWater.initial_infiltration_rate_type == None,
            models.GroundWater.initial_infiltration_rate_file != None
        ),
        message="GroundWater.phreatic_storage_capacity_type cannot be null when "
                "GroundWater.phreatic_storage_capacity_file is not null"
    ),
    QueryCheck(
        column=models.Interflow.porosity,
        invalid=Query(models.Interflow).filter(
            models.Interflow.porosity == None,
            models.Interflow.interflow_type != constants.InterflowType.NO_INTERLFOW
        ),
        message=f"Interflow.porosity cannot be null when "
                f"Interflow.interflow_type != {constants.InterflowType.NO_INTERLFOW}"
    ),
    QueryCheck(
        column=models.Interflow.porosity_layer_thickness,
        invalid=Query(models.Interflow).filter(
            models.Interflow.porosity_layer_thickness == None,
            models.Interflow.interflow_type in [
                constants.InterflowType.LOCAL_DEEPEST_POINT_SCALED_POROSITY,
                constants.InterflowType.GLOBAL_DEEPEST_POINT_SCALED_POROSITY,
            ]
        ),
        message=f"Interflow.porosity_layer_thickness cannot be null when "
                f"Interflow.interflow_type is "
                f"{constants.InterflowType.LOCAL_DEEPEST_POINT_SCALED_POROSITY} or "
                f"{constants.InterflowType.GLOBAL_DEEPEST_POINT_SCALED_POROSITY}"
    ),
    QueryCheck(
        column=models.Interflow.impervious_layer_elevation,
        invalid=Query(models.Interflow).filter(
            models.Interflow.impervious_layer_elevation == None,
            models.Interflow.interflow_type != constants.InterflowType.NO_INTERLFOW
        ),
        message=f"Interflow.impervious_layer_elevation cannot be null when "
                f"Interflow.interflow_type is not {constants.InterflowType.NO_INTERLFOW}"  # noqa: E501
    ),
    QueryCheck(
        column=models.Interflow.hydraulic_conductivity,
        invalid=Query(models.Interflow).filter(
            or_(
                models.Interflow.hydraulic_conductivity == None,
                models.Interflow.hydraulic_conductivity_file == None,
            ),
            models.Interflow.interflow_type != constants.InterflowType.NO_INTERLFOW
        ),
        message=f"Interflow.hydraulic_conductivity cannot be null or "
                f"Interflow.hydraulic_conductivity_file cannot be null when "
                f"Interflow.interflow_type != {constants.InterflowType.NO_INTERLFOW}"
    ),
    QueryCheck(
        column=models.Channel.calculation_type,
        invalid=Query(models.Channel).filter(
            models.Channel.calculation_type.in_([
                constants.CalculationType.EMBEDDED,
                constants.CalculationType.CONNECTED,
                constants.CalculationType.DOUBLE_CONNECTED
            ]),
            models.GlobalSetting.dem_file == None
        ),
        message=f"Channel.calculation_type cannot be "
                f"{constants.CalculationType.EMBEDDED} or"
                f"{constants.CalculationType.CONNECTED} or "
                f"{constants.CalculationType.DOUBLE_CONNECTED} when "
                f"GlobalSetting.dem_file is null"
    ),
    QueryCheck(
        column=models.Pumpstation.lower_stop_level,
        invalid=Query(models.Pumpstation).join(
            models.ConnectionNode,
            models.Pumpstation.connection_node_start_id == models.ConnectionNode.id
        ).join(
            models.Manhole
        ).filter(
            models.Pumpstation.lower_stop_level <= models.Manhole.bottom_level,
        ),
        message="Pumpstation.lower_stop_level should be higher than "
                "Manhole.bottom_level"
    ),
    QueryCheck(
        column=models.Pumpstation.lower_stop_level,
        invalid=Query(models.Pumpstation).join(
            models.ConnectionNode,
            models.Pumpstation.connection_node_end_id == models.ConnectionNode.id
        ).join(
            models.Manhole
        ).filter(
            models.Pumpstation.lower_stop_level <= models.Manhole.bottom_level,
        ),
        message="Pumpstation.lower_stop_level should be higher than "
                "Manhole.bottom_level"
    ),
    QueryCheck(
        column=models.Pipe.invert_level_end_point,
        invalid=Query(models.Pipe).join(
            models.ConnectionNode,
            models.Pipe.connection_node_end_id == models.ConnectionNode.id
        ).join(
            models.Manhole
        ).filter(
            models.Pipe.invert_level_end_point < models.Manhole.bottom_level,
        ),
        message="Pipe.invert_level_end_point should be higher than or equal to "
                "Manhole.bottom_level"
    ),
    QueryCheck(
        column=models.Pipe.invert_level_start_point,
        invalid=Query(models.Pipe).join(
            models.ConnectionNode,
            models.Pipe.connection_node_start_id == models.ConnectionNode.id
        ).join(
            models.Manhole
        ).filter(
            models.Pipe.invert_level_start_point < models.Manhole.bottom_level,  # noqa: E501
        ),
        message="Pipe.invert_level_start_point should be higher than or equal to "
                "Manhole.bottom_level"
    ),
    QueryCheck(
        column=models.Manhole.bottom_level,
        invalid=Query(models.Manhole).filter(
            models.Manhole.drain_level < models.Manhole.bottom_level,
            models.Manhole.calculation_type == constants.CalculationTypeNode.CONNECTED
        ),
        message="Manhole.drain_level >= Manhole.bottom_level when "
                "Manhole.calculation_type is CONNECTED"
    ),
    QueryCheck(
        column=models.Manhole.drain_level,
        invalid=Query(models.Manhole).filter(
            models.Manhole.calculation_type == constants.CalculationTypeNode.CONNECTED,
            models.Manhole.drain_level == None
        ),
        message="Manhole.drain_level cannot be null when Manhole.calculation_type is "
                "CONNECTED"
    ),
    QueryCheck(
        column=models.GlobalSetting.maximum_sim_time_step,
        invalid=Query(models.GlobalSetting).filter(
            models.GlobalSetting.timestep_plus == True,
            models.GlobalSetting.maximum_sim_time_step == None
        ),
        message="GlobalSettings.maximum_sim_time_step cannot be null when "
                "GlobalSettings.timestep_plus is True"
    ),
    QueryCheck(
        column=models.GlobalSetting.use_1d_flow,
        invalid=Query(models.GlobalSetting).filter(
            models.GlobalSetting.use_1d_flow == False,
            Query(func.count(models.ConnectionNode.id) > 0).label("1d_count")
        ),
        message="GlobalSettings.use_1d_flow must be set to True when there are 1d "
                "elements in the model"
    ),
    QueryCheck(
        column=models.GlobalSetting.start_time,
        invalid=Query(models.GlobalSetting).filter(
            func.date(models.GlobalSetting.start_time) == None,
            models.GlobalSetting.start_time != None
        ),
        message="GlobalSettings.start_time is an invalid, make sure it has the "
                "following format: 'HH:MM:SS'"
    ),
    QueryCheck(
        column=models.GlobalSetting.start_date,
        invalid=Query(models.GlobalSetting).filter(
            func.date(models.GlobalSetting.start_date) == None,
            models.GlobalSetting.start_date != None
        ),
        message="GlobalSettings.start_date is an invalid, make sure it has the "
                "following format: 'YYYY-MM-DD'"
    ),
    QueryCheck(
        column=models.Channel.id,
        invalid=Query(models.Channel).filter(
            geo_func.ST_Length(
                geo_func.ST_Transform(
                    models.Channel.the_geom,
                    Query(models.GlobalSetting.epsg_code).limit(1).label("epsg_code")
                )
            ) < 0.05
        ),
        message="Length of the the_geom is too short, should be at least 0.05m"
    ),
    QueryCheck(
        column=models.Culvert.id,
        invalid=Query(models.Culvert).filter(
            geo_func.ST_Length(
                geo_func.ST_Transform(
                    models.Culvert.the_geom,
                    Query(models.GlobalSetting.epsg_code).limit(1).label("epsg_code")
                )
            ) < 0.05
        ),
        message="Length of the the_geom is too short, should be at least 0.05m"
    ),
    ConnectionNodesLength(
        column=models.Pipe.id,
        start_node=models.Pipe.connection_node_start,
        end_node=models.Pipe.connection_node_end,
        min_distance=0.05
    ),
    ConnectionNodesLength(
        column=models.Weir.id,
        start_node=models.Weir.connection_node_start,
        end_node=models.Weir.connection_node_end,
        min_distance=0.05
    ),
    ConnectionNodesLength(
        column=models.Orifice.id,
        start_node=models.Orifice.connection_node_start,
        end_node=models.Orifice.connection_node_end,
        min_distance=0.05
    ),
]


class Config:
    """Collection of checks

    Some checks are generated by a factory. These are usually very generic
    checks which apply to many columns, such as foreign keys."""

    def __init__(self, models):
        self.models = models
        self.checks = []
        self.generate_checks()

    def generate_checks(self):
        FOREIGN_KEY_CHECKS = []
        UNIQUE_CHECKS = []
        NOT_NULL_CHECKS = []
        INVALID_TYPE_CHECKS = []
        INVALID_GEOMETRY_CHECKS = []
        INVALID_GEOMETRY_TYPE_CHECKS = []
        INVALID_ENUM_CHECKS = []
        # Call the check factories:
        for model in self.models:
            FOREIGN_KEY_CHECKS += generate_foreign_key_checks(model.__table__)
            UNIQUE_CHECKS += generate_unique_checks(model.__table__)
            NOT_NULL_CHECKS += generate_not_null_checks(model.__table__)
            INVALID_TYPE_CHECKS += generate_type_checks(model.__table__)
            INVALID_GEOMETRY_CHECKS += generate_geometry_checks(model.__table__)
            INVALID_GEOMETRY_TYPE_CHECKS += generate_geometry_type_checks(model.__table__)  # noqa: E501
            INVALID_ENUM_CHECKS += generate_enum_checks(model.__table__)

        self.checks += FOREIGN_KEY_CHECKS
        self.checks += UNIQUE_CHECKS
        self.checks += NOT_NULL_CHECKS
        self.checks += INVALID_TYPE_CHECKS
        self.checks += INVALID_GEOMETRY_CHECKS
        self.checks += INVALID_GEOMETRY_TYPE_CHECKS
        self.checks += INVALID_ENUM_CHECKS
        self.checks += OTHER_CHECKS
        self.checks += TIMESERIES_CHECKS
        self.checks += RANGE_CHECKS
        self.checks += CONDITIONAL_CHECKS
        return None
